package com.gnss.keepmeclose;

import android.Manifest;
import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothSocket;
import android.content.DialogInterface;
import android.content.Intent;
import android.content.pm.PackageManager;
import android.graphics.Bitmap;
import android.graphics.Matrix;
import android.graphics.drawable.BitmapDrawable;
import android.hardware.Sensor;
import android.hardware.SensorEvent;
import android.hardware.SensorEventListener;
import android.hardware.SensorManager;
import android.location.Criteria;
import android.location.Location;
import android.location.LocationListener;
import android.location.LocationManager;
import android.location.LocationProvider;
import android.media.MediaPlayer;
import android.os.Handler;
import android.os.Message;
import android.support.v4.app.ActivityCompat;
import android.support.v4.app.FragmentActivity;
import android.os.Bundle;
import android.support.v7.app.AlertDialog;
import android.util.Log;
import android.view.animation.Animation;
import android.view.animation.AnimationUtils;
import android.widget.ImageView;
import android.widget.LinearLayout;
import android.widget.TextView;
import android.widget.Toast;
import com.google.android.gms.maps.CameraUpdateFactory;
import com.google.android.gms.maps.GoogleMap;
import com.google.android.gms.maps.OnMapReadyCallback;
import com.google.android.gms.maps.SupportMapFragment;
import com.google.android.gms.maps.model.LatLng;
import com.google.android.gms.maps.model.Marker;
import com.google.android.gms.maps.model.MarkerOptions;
import java.io.InputStream;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.UUID;
import java.io.IOException;

public class MapsActivity extends FragmentActivity implements OnMapReadyCallback {
    static final int MSG_TIMER_TICK = 100;
    static final int MSG_BT_GOT_DATA = 201;
    static final int MSG_UPDATE_LOCATION = 202;

    private enum GGA_STATUS {
        NON_FIX,
        DGPS_FIX,
        GPS_FIX,
        FIXING,
    }
    private TextView myLocation;
    private TextView urLocation;
    private TextView tvDistance;
    private ImageView ivDir;
    private LinearLayout ll = null;

    Location myLoc, devLoc;
    private GoogleMap mMap;
    private String macAddress = "";
    private Timer timer = new Timer();
    private BluetoothAdapter mBluetoothAdapter = null;
    private BluetoothDevice btDevice = null;
    private BluetoothSocket mmSocket = null;;
    private InputStream mmInputStream;
    private boolean stopWorker = false;
    private int readBufferPosition = 0;
    private boolean deviceOutOfRange = false;
    private byte[] readBuffer;
    private Thread workerThread;

    private String[] macList;
    private Marker dvcMarker = null;
    private class BluetoothTask extends TimerTask{
        @Override
        public void run() {
            Message message = new Message();
            message.what = MSG_TIMER_TICK;
            btHandler.sendMessage(message);
        }
    }

    private Handler btHandler = new Handler()  {
        public void handleMessage(Message msg) {
            Bundle b = msg.getData();
            switch (msg.what) {
                case MSG_TIMER_TICK:
                    onTimerTick();
                    break;
                case MSG_BT_GOT_DATA:
                    onBtGotData(b.getString("data"));
                    break;
                case MSG_UPDATE_LOCATION:
                    onUpdateLocation(b.getDouble("Latitude"), b.getDouble("Longitude"));
                    break;
            }
        }
    };

    LocationListener locationListener = new LocationListener() {
        @Override
        public void onLocationChanged(Location loc) {
            // TODO Auto-generated method stub
            Bundle data = new Bundle();

            data.putDouble("Latitude", loc.getLatitude());
            data.putDouble("Longitude", loc.getLongitude());
            data.putDouble("Altitude", loc.getAltitude());

            Message message = new Message();
            message.setData(data);
            message.what = MSG_UPDATE_LOCATION;
            btHandler.sendMessage(message);
        }


        @Override
        public void onProviderDisabled(String provider) {
            // TODO Auto-generated method stub
            //定位提供者如果關閉時會回呼，並將關閉的提供者傳至provider字串中
        }

        @Override
        public void onProviderEnabled(String provider) {
            // TODO Auto-generated method stub
        }

        @Override
        public void onStatusChanged(String provider, int status, Bundle extras) {
            // TODO Auto-generated method stub
        }
    };

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_maps);

        if(macAddress.isEmpty()) {
            showDeviceSelect();
        }
        else  {
            openDevice();
        }

        // Obtain the SupportMapFragment and get notified when the map is ready to be used.
        SupportMapFragment mapFragment = (SupportMapFragment) getSupportFragmentManager().
                findFragmentById(R.id.map);
        mapFragment.getMapAsync(this);
        ll = (LinearLayout)findViewById(R.id.linearLayout);
        myLocation = (TextView)ll.findViewById(R.id.myLocation);
        urLocation = (TextView)ll.findViewById(R.id.deviceLocation);
        tvDistance = (TextView)ll.findViewById(R.id.distance);
        ivDir = (ImageView)ll.findViewById(R.id.imageView);
        mBluetoothAdapter = BluetoothAdapter.getDefaultAdapter();

        SensorManager sm = (SensorManager)getSystemService(SENSOR_SERVICE);
        Sensor oSensor = sm.getDefaultSensor(Sensor.TYPE_ORIENTATION);
        sm.registerListener(myListener, oSensor, SensorManager.SENSOR_DELAY_NORMAL);
    }

    float oriSensorDegree = Float.MAX_VALUE;
    final SensorEventListener myListener = new SensorEventListener() {
        public void onSensorChanged(SensorEvent sensorEvent) {
            if (sensorEvent.sensor.getType() == Sensor.TYPE_ORIENTATION) {
                oriSensorDegree = sensorEvent.values[0];
            }
        }
        public void onAccuracyChanged(Sensor sensor, int accuracy) {}
    };
    boolean  isResumed = false;
    @Override
    protected void onPause() {
        super.onPause();
        isResumed = false;
        try  {
            closeBT();
        }
        catch (IOException ex) { }

    }

    @Override
    protected void onResume() {
        super.onResume();
        isResumed = true;
    }

    @Override
    public void onMapReady(GoogleMap googleMap) {
        mMap = googleMap;
        mMap.setMapType(GoogleMap.MAP_TYPE_NORMAL); //map mode
        if (ActivityCompat.checkSelfPermission(this, Manifest.permission.ACCESS_FINE_LOCATION) == PackageManager.PERMISSION_GRANTED &&
                ActivityCompat.checkSelfPermission(this, Manifest.permission.ACCESS_COARSE_LOCATION) == PackageManager.PERMISSION_GRANTED) {
            mMap.setMyLocationEnabled(true);
        }

        LocationManager locationManager = (LocationManager)getSystemService(this.LOCATION_SERVICE);
        locationManager.requestLocationUpdates(LocationManager.GPS_PROVIDER, 1000, 0, locationListener);

        Criteria criteria = new Criteria();
        criteria.setAccuracy(Criteria.ACCURACY_FINE);
        String best = locationManager.getBestProvider(criteria, true);
        Location location = locationManager.getLastKnownLocation(best);
        if(location != null) {
            LatLng nkut = new LatLng(location.getLatitude(), location.getLongitude());
            mMap.moveCamera(CameraUpdateFactory.newLatLngZoom(nkut, 18.0f));
        }
        else  {
            location = locationManager.getLastKnownLocation(LocationManager.NETWORK_PROVIDER);
            if(location != null) {
                LatLng nkut = new LatLng(location.getLatitude(), location.getLongitude());
                mMap.moveCamera(CameraUpdateFactory.newLatLngZoom(nkut, 18.0f));
            }
        }
    }

    private void onTimerTick() {
        if(btDevice == null) {
            btDevice = mBluetoothAdapter.getRemoteDevice(macAddress);
            try {
                findBT();
                openBT();
            }
            catch (IOException ex) { }
        }

        if(deviceOutOfRange)
        {
            if(mmSocket == null) {
                try {
                    openBT();
                }
                catch (IOException ex) { }
            }
            else {
                try {
                    closeBT();
                }
                catch (IOException ex) { }
            }
        }

        if(!deviceOutOfRange && --deviceCounter == 0)  {
            setDeviceOutOfRange(true);
            devLoc = null;
            if(dvcMarker != null)
                dvcMarker.remove();
        }
        updatePanel();
    }

    private double parseLonLat(String coorString, String dirChar)  {
        int dd;
        float mm;
        double tmp, lonLat;

        tmp = Double.parseDouble(coorString);
        dd = (int)(tmp / 100);
        mm = (float)(tmp - dd * 100);
        lonLat = dd + mm/60;

        if (dirChar.equals("W") || dirChar.equals("S"))
            lonLat = 0 - lonLat;
        return lonLat;
    }

    final int deviceCount = 3;
    int deviceCounter = deviceCount;
    private void setDeviceOutOfRange(boolean b)  {
        if(b && deviceOutOfRange == false)
        {   //First time out of range, close connection
            try  {
                closeBT();
            }
            catch (IOException ex) { }
        }

        if(b)  {
            deviceOutOfRange = true;
        }
        else  {
            deviceCounter = deviceCount;
            deviceOutOfRange = false;
        }
    }

    private void onBtGotData(String str)  {
        try {
            String temp = str.substring(0, str.indexOf('*'));
            String[] nmeaArray = temp.split(",", -1);
            setDeviceOutOfRange(false);
            if (nmeaArray[1].length() == 0) return;
    //        int hour = Integer.parseInt(nmeaArray[1].substring(0, 2));
    //        int minute = Integer.parseInt(nmeaArray[1].substring(2, 4));
    //        float second = Float.parseFloat(nmeaArray[1].substring(4, nmeaArray[1].length()));

            double lat = 0, lon = 0;
            if (nmeaArray[2].length() == 0 || nmeaArray[3].length() == 0)
                lat = 0;
            else
                lat = parseLonLat(nmeaArray[2], nmeaArray[3]);
            if (nmeaArray[4].length() == 0 || nmeaArray[5].length() == 0)
                lon = 0;
            else
                lon = parseLonLat(nmeaArray[4], nmeaArray[5]);
            GGA_STATUS gga_status = GGA_STATUS.NON_FIX;
            if (nmeaArray[6].equals("0"))
                gga_status = GGA_STATUS.NON_FIX;
            else if (nmeaArray[6].equals("1"))
                gga_status = GGA_STATUS.DGPS_FIX;
            else if (nmeaArray[6].equals("2"))
                gga_status = GGA_STATUS.GPS_FIX;
            else if (nmeaArray[6].equals("3"))
                gga_status = GGA_STATUS.FIXING;

            if(gga_status != GGA_STATUS.NON_FIX)  {
                String coorStr = String.format("     %f, %f", lat, lon);
                //if(urLocation.getText().toString().compareTo(coorStr) != 0) {
                    urLocation.setText(coorStr);
                    Animation am = AnimationUtils.loadAnimation(this, R.anim.scale);
                    urLocation.startAnimation(am);

                    if(dvcMarker != null) {
                        dvcMarker.remove();
                    }

                    MarkerOptions opt = new MarkerOptions();
                    opt.position(new LatLng(lat, lon));
                    opt.title(coorStr);
                    dvcMarker = mMap.addMarker(opt);
                //}

                if(devLoc == null) {
                    devLoc = new Location("");
                }
                devLoc.setLatitude(lat);
                devLoc.setLongitude(lon);
                devLoc.setAltitude(0);
                updatePanel();
            }
            else {
                String coorStr = String.format("No Position Fix.");
                //if(urLocation.getText().toString().compareTo(coorStr) != 0) {
                    urLocation.setText(coorStr);
                    Animation am = AnimationUtils.loadAnimation(this, R.anim.scale);
                    urLocation.startAnimation(am);

                    if (dvcMarker != null) {
                        dvcMarker.remove();
                    }
                //}
            }
        }
        catch(Exception e) {
            Log.d("CloseToMe", str);
        }

    }

    private void onUpdateLocation(double latitude, double longitude)  {
        String coorStr = String.format("     %f, %f", latitude, longitude);
        if(myLocation.getText().toString().compareTo(coorStr) != 0)  {
            myLocation.setText(coorStr);
            Animation am = AnimationUtils.loadAnimation(this, R.anim.scale);
            myLocation.startAnimation(am);
        }
        if(myLoc == null)  {
            myLoc = new Location("");
        }
        myLoc.setLatitude(latitude);
        myLoc.setLongitude(longitude);
        myLoc.setAltitude(0);
        updatePanel();
    }

    private double gps2d(double lat_a, double lng_a, double lat_b, double lng_b) {
        double d;
        lat_a = lat_a * Math.PI / 180;
        lng_a = lng_a * Math.PI / 180;
        lat_b = lat_b * Math.PI / 180;
        lng_b = lng_b * Math.PI / 180;

        d = Math.sin(lat_a) * Math.sin(lat_b) + Math.cos(lat_a) * Math.cos(lat_b) * Math.cos(lng_b - lng_a);
        d = Math.sqrt(1 - d * d);
        d = Math.cos(lat_b) * Math.sin(lng_b - lng_a) / d;
        d = Math.asin(d) * 180 / Math.PI;
        return d;
    }

    private int calcGoogleMapZoomScale(double d)  {
        final int minScale = 20;
        if(d <= 10.0)
            return minScale;
        int zs = minScale - (int)((Math.log(d) - 1) / (Math.log(20) - 1));
        return (zs < 0) ? 0 : zs;
    }

    int lastScale = 0;
    MediaPlayer mp = null;
    private void updatePanel()  {
        if(myLoc != null && devLoc !=null) {
            double d = gps2d(myLoc.getLatitude(), myLoc.getLongitude(),
                    devLoc.getLatitude(), devLoc.getLongitude());
            double dis = GetDistance(myLoc.getLongitude(), myLoc.getLatitude(),
                    devLoc.getLongitude(), devLoc.getLatitude());

            tvDistance.setText(String.format("%1.1fm", dis));
            setRotation((float)(d - oriSensorDegree));

            int scale = calcGoogleMapZoomScale(dis);
            if (scale != lastScale)  {
                mMap.moveCamera(CameraUpdateFactory.zoomTo(scale));
                lastScale = scale;
                mMap.moveCamera(CameraUpdateFactory.newLatLng(
                        new LatLng(myLoc.getLatitude(), myLoc.getLongitude())));
            }
        }
        else  {
            if(myLoc == null)  {
                myLocation.setText("No Position Fix.");
            }
            try {
                if (deviceOutOfRange) {
                    urLocation.setText("Device out of range!");
                    if (mp == null) {
                        mp = MediaPlayer.create(this, R.raw.bb);
                        mp.setLooping(true);
                        mp.start();
                    }
                    if (!mp.isPlaying()) {
                        mp.release();
                        mp = null;
                    }
                }
                else {
                    if (mp != null && mp.isPlaying()) {
                        mp.stop();
                        mp.release();
                        mp = null;
                    }
                }
            }
            catch (Exception e)
            {
                Log.d("ClostToMe", e.toString());
            }
        }
    }

    private void findBT()  {
        mBluetoothAdapter = BluetoothAdapter.getDefaultAdapter();
        if(mBluetoothAdapter == null)  {
            Toast.makeText(MapsActivity.this,
                    "No bluetooth adapter available!", Toast.LENGTH_SHORT).show();
        }

        if(!mBluetoothAdapter.isEnabled())  {
            Intent enableBluetooth = new Intent(BluetoothAdapter.ACTION_REQUEST_ENABLE);
            startActivityForResult(enableBluetooth, 0);
        }

        Set<BluetoothDevice> pairedDevices = mBluetoothAdapter.getBondedDevices();
        if(pairedDevices.size() > 0)  {
            for(BluetoothDevice device : pairedDevices)  {
                if(device.getAddress().equals(macAddress))  {
                    btDevice = device;
                    break;
                }
            }
        }
    }

    private void openBT() throws IOException  {
        //Standard SerialPortService ID
        UUID uuid = UUID.fromString("00001101-0000-1000-8000-00805f9b34fb");
        mmSocket = btDevice.createInsecureRfcommSocketToServiceRecord(uuid);
        mBluetoothAdapter.cancelDiscovery();
        mmSocket.connect();
        mmInputStream = mmSocket.getInputStream();
        beginListenForData();
    }

    void closeBT() throws IOException  {
        stopWorker = true;
        if(mmInputStream != null) {
            mmInputStream.close();
            mmInputStream = null;
        }
        if(mmSocket != null) {
            mmSocket.close();
            mmSocket = null;
        }
    }

    private void beginListenForData()  {
        final byte delimiter = 10; //This is the ASCII code for a newline character
        stopWorker = false;
        readBufferPosition = 0;
        readBuffer = new byte[1024];
        workerThread = new Thread(new Runnable()  {
            public void run()  {
            while(!Thread.currentThread().isInterrupted() && !stopWorker)  {
                try  {
                    int bytesAvailable = mmInputStream.available();
                    if(bytesAvailable > 0)  {
                        byte[] packetBytes = new byte[bytesAvailable];
                        mmInputStream.read(packetBytes);
                        for(int i=0;i<bytesAvailable;i++)  {
                            byte b = packetBytes[i];
                            if(b == delimiter)  {
                                byte[] encodedBytes = new byte[readBufferPosition];
                                if(readBuffer[0] == '$' && readBuffer[3] == 'G' &&
                                        readBuffer[4] == 'G' && readBuffer[5] == 'A') {
                                    //Parsing NMEA GGA token.
                                    System.arraycopy(readBuffer, 0,
                                            encodedBytes, 0, encodedBytes.length);
                                    Bundle data = new Bundle();
                                    data.putString("data",
                                            new String(encodedBytes, "US-ASCII"));
                                    readBufferPosition = 0;

                                    Message message = new Message();
                                    message.setData(data);
                                    message.what = MSG_BT_GOT_DATA;
                                    btHandler.sendMessage(message);
                                }
                                else  {
                                    readBufferPosition = 0;
                                }
                            }
                            else  {
                                readBuffer[readBufferPosition++] = b;
                            }
                        }
                    }
                    else  {
                        try  {
                            Thread.sleep(100);
                        }
                        catch (InterruptedException e)  {
                        }
                    }
                }
                catch (IOException ex)  {
                    stopWorker = true;
                }
            }
            }
        });
        workerThread.start();
    }

    private void openDevice()  {
        if (!BluetoothAdapter.checkBluetoothAddress(macAddress))  {
            Toast.makeText(MapsActivity.this,
                    "Invalid Bluetooth MAC Address!", Toast.LENGTH_SHORT).show();
        }
        else  {
            timer.scheduleAtFixedRate(new BluetoothTask(), 0, 1000L);
        }
    }

    private void showDeviceSelect()  {
        final String[] deviceList = bluetoothDevicesList();
        AlertDialog.Builder builder = new AlertDialog.Builder(MapsActivity.this);
        builder.setTitle("Select a bluetooth device").
                setPositiveButton("Cancel", new DialogInterface.OnClickListener() {
                    public void onClick(DialogInterface dialog, int which) {
                        macAddress = "";
                    }
                });

        builder.setItems(deviceList, new DialogInterface.OnClickListener() {
            public void onClick(DialogInterface dialog, int which) {
                macAddress = macList[which];
                openDevice();
            }
        });
        builder.show();
    }

    private String[] bluetoothDevicesList()  {
        BluetoothAdapter bt = BluetoothAdapter.getDefaultAdapter();
        if(bt == null)  {
            Toast.makeText(this, "Please turn on Bluetooth", Toast.LENGTH_SHORT).show();
            return null;
        }

        Set<BluetoothDevice> pairedDevices = bt.getBondedDevices();
        CharSequence[] entries = new CharSequence[pairedDevices.size()];
        CharSequence[] entryValues = new CharSequence[pairedDevices.size()];
        int i = 0;
        for (BluetoothDevice dev : pairedDevices)  {
            entries[i] = dev.getName();
            entryValues[i] = dev.getAddress();
            i++;
        }

        if(i > 0)  {
            String[] nameList = new String[i];
            macList = new String[i];
            for(;i != 0; --i)  {
                nameList[i - 1] = entries[i -  1].toString();
                macList[i - 1] = entryValues[i -  1].toString();
            }
            return nameList;
        }
        return null;
    }

    public static double GetDistance(double wd1, double jd1, double wd2, double jd2)  {
        double x, y;
        double R = 6371229;
        x = (jd2 - jd1) * Math.PI * R * Math.cos( ( (wd1 + wd2) / 2) * Math.PI / 180) / 180;
        y = (wd2 - wd1) * Math.PI * R / 180;
        return Math.hypot(x, y);
    }

    private Matrix matrix = new Matrix();
    private void setRotation(float angle)  {
        angle = (angle + 360) % 360;

        Bitmap bitmap =
                ((BitmapDrawable)(getResources().getDrawable(R.drawable.direction))).getBitmap();
        matrix.setRotate(angle);
        Bitmap bmp = Bitmap.createBitmap(bitmap, 0, 0,
                bitmap.getWidth(), bitmap.getHeight(), matrix, true);
        ivDir.setImageBitmap(bmp);
    }
}
